"""Create online experiments"""

icon = "preferences-desktop"
label = "OSWeb and JATOS control panel"
settings = {
  "oswebext_include_context": True,
  "oswebext_intro_click": True,
  "oswebext_fullscreen": False,
  "oswebext_full_background_color": False,
  "oswebext_bypass_linter": False,
  "oswebext_subject_nrs": "0,1",
  "oswebext_welcome_text": "",
  "oswebext_external_js": "",
  "oswebext_jatos_url": "https://jatos.mindprobe.eu",
  "oswebext_jatos_api_token": "",
  "oswebext_jatos_ignore_conflicts": False
}
menu = {
  "index": -1,
  "separator_after": False,
  "separator_before": False,
  "submenu": "Tools"
}



"""This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from collections import namedtuple
from pathlib import Path
import json
import requests
import hashlib
import zlib
from libopensesame.oslogging import oslogger
from . import convert, sync
from .oswebexceptions import VersionInfoDownloadError


VersionComparison = namedtuple(
    'VersionComparison', ('unmodified', 'modified', 'conflicting', 'added',
                          'deleted'))


def compare_version_info(jatos_info, older_info, newer_info):
    """Compares two version infos to see whether the experiment corresponding
    to older_info can be safely updated to newer_info. Returns a 
    VersionComparison object that contains lists of files that:
    
    - have been unmodified in older_info relative to newer_info (unmodified)
    - have been modified in older_info relative to newer_info but can be
      fast-forwarded without conflict (modified)
    - have been modified in older_info relative to newer_info and cannot be
      fast-forwarded without conflict (conflicting)
    - have been added in newer_info relative to older_info (added)
    - have been deleted in newer_info relative to older_info (deleted)
    """

    def get_src_file_history(newer_versions, file):
        return [v[file] for v in newer_versions if file in v]

    newer_versions = newer_info[jatos_info.url]['versions']
    older_versions = older_info[jatos_info.url]['versions']
    newer_latest = newer_versions[-1]
    older_latest = older_versions[-1]
    newer_files = set(newer_latest.keys())
    older_files = set(older_latest.keys())
    all_files = newer_files & older_files
    # A file is unmodified if the latest versions in the newer and older info
    # are the same
    unmodified = set(file for file in all_files
                     if newer_latest[file] == older_latest[file])
    # A file is modified if the latest versions in the newer and older info
    # don't match, but the old version is present in the history of the newer
    # info
    modified = set(file for file in all_files
                   if newer_latest[file] != older_latest[file] 
                   and older_latest[file] in get_src_file_history(
                       newer_versions, file))
    # A file is conflicting if trhe latest versions in the newer and older info
    # don't match, and the old version is not in th ehistory of the newer info
    conflicting = set(file for file in all_files
                      if newer_latest[file] != older_latest[file]
                      and older_latest[file] not in get_src_file_history(
                          newer_versions, file))
    added = newer_files - older_files
    deleted = older_files - newer_files
    return VersionComparison(unmodified, modified, conflicting, added, deleted)


def download_version_info(exp, jatos_info):
    """Downloads version info (version_info.json) from JATOS and returns it as
    a dict. If the study doesn't exist or if it does exist but does not contain
    any version info, None is returned.
    """
    headers = {'accept': 'application/json',
               'Authorization': f'Bearer {jatos_info.token}'}
    try:
        response = requests.get(
            f'{jatos_info.url}/jatos/api/v1/studies/{exp.var.jatos_uuid}'
            f'/assets/version_info.json',
            headers=headers)
    except Exception as e:
        raise VersionInfoDownloadError(
            f'Failed to connect to JATOS server: \n\n{e}')
    if response.status_code == 200:
        oslogger.debug('version info succesfully downloaded')
        try:
            return response.json()
        except Exception as e:
            raise VersionInfoDownloadError(
                f'Failed to parse version info: \n\n{e}')
    oslogger.warning('failed to download version info')


def update_version_info(exp, asset_list, jatos_info):
    """Updates version_info.json in the file pool (creating it if it doesn't
    exist yet) and return the version info as a dict.
    """
    exp = convert.as_jatos_exp(exp)
    folder = Path(exp.pool.folder())
    version_info_path = folder / 'version_info.json'
    if version_info_path.exists():
        version_info = json.loads(version_info_path.read_text())
    else:
        oslogger.debug('initializing version_info.json')
        version_info = {}
    # Initialize mandatory fields if they don't exist yet
    version_info.setdefault(jatos_info.url, {})
    version_info[jatos_info.url].setdefault('uuid', exp.var.jatos_uuid)
    version_info[jatos_info.url].setdefault('versions', [])
    # Compute the version dict for the current state of the files in the folder
    current_version = {}
    for src, tgt in asset_list:
        if isinstance(src, Path):
            if not src.exists():
                oslogger.debug(f'asset does not exist: {src}')
                continue
            src = src.read_bytes()
        else:
            src = src.encode('utf-8')
        sha256 = adler32(src)
        current_version[tgt] = sha256
        oslogger.debug(f'{tgt}: {sha256}')
    # Append the current version to the list of versions
    version_info[jatos_info.url]['versions'].append(current_version)
    # Save back to the file and return
    version_info_path.write_text(json.dumps(version_info))
    return version_info


def extract_version_info(exp):
    path = Path(convert.as_jatos_exp(exp).pool.folder()) / 'version_info.json'
    return json.loads(path.read_text())


def adler32(data):
    """Compute the Adler32 checksum of the file specified by `path`."""
    return zlib.adler32(data) & 0xffffffff  # Return a 32-bit integer<


"""This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import OSException


class OSWebException(OSException):
    def __str__(self):
        return f'{self.title()}\n\n{self._msg}'

    def markdown(self):
        return f'# {self.title()}\n\n{self._msg}\n\n{self._read_more}'


class PythonToJavaScriptError(OSWebException):
    """A `PythonToJavaScriptError` is raised when a Python expression could
    not be converted to JavaScript. This may happen when a Python expression is
    invalid or when there is no equivalent of a Python expression in
    JavaScript. To resolve the error, locate and fix the Python expression
    mentioned in the error message.
    """
        
        
class OSWebCompatibilityError(OSWebException):
    """An `OSWebCompatibilityError` is raised when an experiment contains
    functionality that is not supported by OSWeb. This may happen when the
    experiment contains unsupported items, such as Python `inline_script` 
    items. An overview of supported functionality can be found on the
    documentation site. The compatibility check can be disabled (at your own 
    risk!) in the OSWeb control panel.
    """


class JZIPDownloadError(OSWebException):
    """A `JZIPDownloadError` is raised when an experiment could not be
    downloaded from a JATOS server. This may happen when the JATOS server or
    API token are incorrect. This may also happen when there is a problem with
    the internet connection. Finally, this may also happen when the experiment
    that you are trying to open is not in the correct format, for example when
    it uses jsPsych, lab.js, or an older version of OSWeb.
    """


class VersionInfoDownloadError(OSWebException):
    """A `VersionInfoDownloadError` is raised when version info could not be
    downloaded from a JATOS server. This may happen when the JATOS server or
    API token are incorrect. This may also happen when there is a problem with
    the internet connection. Finally, this may also happen when the experiment
    that you are trying to open is not in the correct format, for example when
    it uses jsPsych, lab.js, or an older version of OSWeb.
    """


class JZIPUploadError(OSWebException):
    """A `JZIPUploadError` is raised when an experiment could not be 
    uploaded to a JATOS server. This may happen when the JATOS server or API 
    token are incorrect. This may also happen when there is a problem with the
    internet connection.
    """


class VersionConflict(OSWebException):
    """A `VersionConflict` is raised when trying to update a remote experiment
    that cannot be safely updated to the current experiment. You can work 
    around this in two ways: 
    
    1. Reset the JATOS UUID to reset the link between the remote experiment and
    the current experiment. This will upload the current experiment as a new
    remote experiment.
    2. Enable 'Ignore conflicts'. This will forcibly update the remote 
    experiment, overwriting conflicting files.
    
    These options are available in the OSWeb and JATOS control panel.
    """


class UnsupportedJZIP(OSWebException):
    """An `UnsupportedJZIP` is raised when opening an unsupported JATOS Study
    Archive (.jzip). This may happen when the study was created with an older
    version of OSWeb or with other software. This may also happen when the
    jzip file is corruped.
    """


class ListRemoteError(OSWebException):
    """A `ListRemoteError` is raised when something goes wrong while trying to
    list all (remote) experiments on a JATOS server. This may happen when the
    JATOS server or API token are incorrect. This may also happen when there is
    a problem with the internet connection.
    """



"""This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from collections import namedtuple
from pathlib import Path
import requests
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
import tempfile
import os
from libopensesame.oslogging import oslogger
from . import version_info
from . import convert
from .oswebexceptions import JZIPDownloadError, JZIPUploadError, \
    ListRemoteError, VersionConflict


JatosInfo = namedtuple('JatosInfo', ('url', 'token'))


def download_jzip(uuid, jatos_info, callback=None, jzip_path=None):
    """Downloads a JZIP file from a JATOS server and saves it to a specified 
    location.

    This function sends a GET request to the JATOS server to download a JZIP 
    file  identified by a specific UUID. The file is saved to a location 
    specified by `jzip_path`, or to a temporary location if no path is 
    provided. The function logs the download  process and raises a 
    `JZIPDownloadError` if the server responds with a non-200 status code.

    Parameters
    ----------
    uuid : str
        The unique identifier of the study on the JATOS server.
    jatos_info : JatosInfo
    callback : callable, optional
        A function that takes a MultipartEncoderMonitor as a single argument.
        This function can be used to monitor upload progress.    
    jzip_path : str, optional
        The path where the JZIP file will be saved. If not provided, the file 
        is saved to a temporary location.

    Returns
    -------
    pathlib.Path
        The path where the JZIP file has been saved.

    Raises
    ------
    JZIPDownloadError
        If the response from the server is not a 200 status code.
    """
    headers = {'accept': 'application/json',
               'Authorization': f'Bearer {jatos_info.token}'}
    try:
        response = requests.get(
            f'{jatos_info.url}/jatos/api/v1/studies/{uuid}', headers=headers,
            stream=True)
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        raise JZIPDownloadError(f'HTTP error occurred: {e}')
    except Exception as e:
        raise JZIPDownloadError(f'Failed to connect to JATOS server: \n\n{e}')

    if jzip_path is None:
        jzip_path = Path(tempfile.NamedTemporaryFile(delete=False).name)
    else:
        jzip_path = Path(jzip_path)
    try:
        with jzip_path.open('wb') as f:
            total_length = int(response.headers.get('content-length'))
            downloaded = 0
            for chunk in response.iter_content(chunk_size=1024):
                downloaded += len(chunk)
                f.write(chunk)
                if callback is not None:
                    callback(downloaded, total_length)
    except Exception as e:
        raise JZIPDownloadError(f'Failed to stream from JATOS server: \n\n{e}')
    return jzip_path


def list_remote(jatos_info):
    """List the properties of remote studies from a JATOS server.
    
    Parameters
    ----------
    jatos_info : JatosInfo
    
    Returns
    -------
    study_list : list
        A list of dictionaries, where each dictionary contains properties
        of a remote study available on the JATOS server. The content of
        each dictionary depends on the properties set in each study on the
        server.
    """
    headers = {'accept': 'application/json',
               'Authorization': f'Bearer {jatos_info.token}'}
    try:
        response = requests.get(
            f'{jatos_info.url}/jatos/api/v1/studies/properties',headers=headers)
    except Exception as e:
        raise ListRemoteError(f'Failed to connect to JATOS server: \n\n{e}')
    if response.status_code == 401:
        raise ListRemoteError(
            'Invalid API token: {jato_info.token} status code 401)')
    try:
        study_list = response.json()['data']
    except Exception as e:
        raise ListRemoteError(
            f'Server does not appear to be a JATOS server: \n\n{e}')
    oslogger.debug(f'listed {len(study_list)} remote studies')
    return study_list


def upload(exp, jatos_info, callback=None, **jzip_kwargs):
    """Upload an experiment to a JATOS server.

    This function converts an experiment into a jzip file, logs the size of the
    file, sends a POST request to the JATOS server to upload the study, and
    finally deletes the local jzip file. The JATOS UUID is returned.

    Parameters
    ----------
    exp : Experiment
        If no UUID was specified in the experiment, the experiment object is 
        modified in place such that a jatos_uuid variabiable is added.
    jatos_info : JatosInfo
    callback : callable, optional
        A function that takes a MultipartEncoderMonitor as a single argument.
        This function can be used to monitor upload progress.
    jzip_kwags : dict, optional
        Parameters that are passed onto convert.exp_to_jzip()
        
    Returns
    -------
    str
        JATOS UUID

    Raises
    ------
    JZIPUploadError
        If the experiment failed to upload for technical reasons
    VersionConflict
        If the experiment failed to upload because of a version conflict
    """
    exp = convert.as_jatos_exp(exp)
    try:
        path = convert.exp_to_jzip(exp, jatos_info=jatos_info, **jzip_kwargs)
    except VersionConflict:
        raise
    except Exception as e:
        raise JZIPUploadError(f'Failed to connect to server: \n\n{e}')
    oslogger.debug(
        f'jzip exported to {path} ({path.stat().st_size / 1024 ** 2:.2f} Mb)')
    params = {
        'keepProperties': 'false', 
        'keepAssets': 'true', 
        'keepCurrentAssetsName': 'true', 
        'renameAssets': 'true'
    }
    with path.open('rb') as fd:
        encoder = MultipartEncoder(
            {'study': ('filename', fd, 'application/zip')})
        # A callback function that calls the actual user-provided callback
        # with the number of bytes read and the total number of bytes
        def upload_callback(monitor):
            if callback is not None:
                callback(monitor.bytes_read, encoder.len)
        monitor = MultipartEncoderMonitor(encoder, upload_callback)
        headers = {'accept': 'application/json',
                   'Authorization': f'Bearer {jatos_info.token}',
                   'Content-Type': monitor.content_type}
        try:
            response = requests.post(f'{jatos_info.url}/jatos/api/v1/study',
                                     headers=headers,
                                     data=monitor)
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise JZIPUploadError(f'HTTP error occurred: {e}')
        except Exception as e:
            raise JZIPUploadError(f'Failed to connect to server: \n\n{e}')
    path.unlink()
    if response.status_code == 200:
        oslogger.debug('succesfully published')
        return exp.var.jatos_uuid
    if response.status_code == 401:
        raise JZIPUploadError(
            'Invalid API token: {jato_info.token} status code 401)')
    if response.status_code == 404:
        raise JZIPUploadError(
            f'Invalid server address: {jatos_info.url} (status code 404)')
    raise JZIPUploadError(
        f'Unknown upload error (status code {response.status_code})')



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.widgets.base_widget import BaseWidget


class OSWebControlPanel(BaseWidget):
    """The OSWeb control panel combines the actions and settings that are also
    available elsewhere in the user interface.
    """
    def __init__(self, parent, settings_widget):
        super().__init__(parent, ui='extensions.oswebext.osweb_control_panel')
        self._settings_widget = settings_widget
        self.ui.layout_vbox.addWidget(settings_widget)
        self.ui.layout_vbox.addStretch()
        self.ui.button_export_html.clicked.connect(self._export_html)
        self.ui.button_export_jzip.clicked.connect(self._export_jzip)
        self.ui.button_import_jzip.clicked.connect(self._import_jzip)
        self.ui.button_open_jatos.clicked.connect(self._open_jatos)
        self.ui.button_publish_jatos.clicked.connect(self._publish_jatos)
        self.ui.button_convert_results.clicked.connect(self._convert_results)
        self.ui.button_experiment_properties.clicked.connect(
            self.tabwidget.open_general)
        self.refresh()
    
    def showEvent(self, event):
        self.refresh()
        
    def refresh(self):
        self._settings_widget.ui.edit_uuid.setText(
            safe_str(self.experiment.var.get('jatos_uuid', '')))
        self._settings_widget.ui.edit_end_redirect_url.setText(
            safe_str(self.experiment.var.get('jatos_end_redirect_url', '')))
        if self.experiment.var.canvas_backend == 'osweb':
            self.ui.button_experiment_properties.setVisible(False)
            self.ui.label_experiment_properties.setVisible(False)
        else:
            self.ui.button_experiment_properties.setVisible(True)
            self.ui.label_experiment_properties.setVisible(True)
            
    def set_jatos_configured(self, configured):
        self.ui.label_jatos_unconfigured.setVisible(not configured)

    def _export_html(self):
        self.extension_manager.fire('osweb_export_html')
        
    def _export_jzip(self):
        self.extension_manager.fire('osweb_export_jzip')
        
    def _import_jzip(self):
        self.extension_manager.fire('osweb_import_jzip')
        
    def _open_jatos(self):
        self.extension_manager.fire('osweb_open_jatos')
        
    def _publish_jatos(self):
        self.extension_manager.fire('osweb_publish_jatos')
        
    def _convert_results(self):
        self.extension_manager.fire('osweb_convert_results')



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.osexpfile import OSExpWriter
from libopensesame.oslogging import oslogger
from metapensiero.pj.__main__ import transform_string as py2js
from .oswebexceptions import PythonToJavaScriptError, OSWebCompatibilityError
import re

RE_VAR = re.compile('(?<!\w)var\.', re.MULTILINE)
RE_RUN_IF = re.compile(r'^\t(?P<cmd>run \w+ .*)$', re.MULTILINE)
RE_SET_COND = re.compile(r'^\t(?P<cmd>set (break_if|condition) .*)$',
                         re.MULTILINE)
RE_SET_COND_MULTILINE = re.compile(
    '__(?P<key>(break_if|condition))__(?P<value>.*?)__end__\n',
    re.MULTILINE | re.DOTALL)
RE_DRAW = re.compile(r'^\t(?P<cmd>draw .*)$', re.MULTILINE)
RE_FSTRING = re.compile(r'(?<!{){(?!{)(?P<expr>.*?)}')
RE_IGNORE_RUN = re.compile('___run__.*?__end__\n', re.MULTILINE | re.DOTALL)
RE_IGNORE_PREPARE = re.compile('___prepare__.*?__end__\n',
                               re.MULTILINE | re.DOTALL)
RE_IGNORE_SCRIPT = re.compile('<script>.*?</script>', re.MULTILINE | re.DOTALL)
RE_STRIP_DISABLED = re.compile(r'^\t(?P<cmd>run \w+ .* disabled)(\n|$)',
                               re.MULTILINE)


class OSWebWriter(OSExpWriter):
    """Translates Python-style conditional expressions and f-string to
    JavaScript before writing the experiment to disk. This means that OSWeb
    doesn't have to do any translating itself.
    """
    
    def transform(self, cond, escape=False):
        py_cond = self._exp.syntax.compile_cond(cond, bytecode=False)
        # JavaScript uses vars as opposed to var for the var_store object
        py_cond = RE_VAR.sub('vars.', py_cond)
        # The last two characters are ;\n, which are not valid in some
        # contexts and we therefore strip them off
        try:
            js_cond = py2js(py_cond)[:-2]
        except Exception as e:
            raise PythonToJavaScriptError(
                f'Failed to convert {cond} to JavaScript')
        if escape:
            js_cond = js_cond.replace('"', r'\"')
        return js_cond
            
    @property
    def create_cmd(self):
        return self._exp.syntax.create_cmd
    
    @property
    def parse_cmd(self):
        return self._exp.syntax.parse_cmd

    @property
    def script(self):
        script = self._exp.to_string()
        # We first remove all the disabled run statements. This way we don't
        # have to implement support for this in OSWeb
        while True:
            m = RE_STRIP_DISABLED.search(script)
            if m is None:
                break
            script = script[:m.start()] + script[m.end():]        
        for m in RE_RUN_IF.finditer(script):
            try:
                _, (item, cond), _ = self.parse_cmd(m.group('cmd'))
            except ValueError:
                # This mainly happens for coroutines, which have a slightly
                # deviant run statement
                raise OSWebCompatibilityError(
                    f'The following OpenSesame script is not supported in '
                    f'OSWeb:\n\n```\n{m.group("cmd")}\n```')
            cond = self.transform(cond)
            cmd = self.create_cmd('run', (item, cond))
            oslogger.debug(f'rewriting {m.group("cmd")} to {cmd}')
            script = script.replace(m.group('cmd'), cmd)
        for m in RE_SET_COND.finditer(script):
            _, (key, cond), _ = self.parse_cmd(m.group('cmd'))
            cond = self.transform(cond)
            cmd = self.create_cmd('set', (key, cond))
            oslogger.debug(f'rewriting {m.group("cmd")} to {cmd}')
            script = script.replace(m.group('cmd'), cmd)
        for m in RE_SET_COND_MULTILINE.finditer(script):
            key = m.group('key')
            value = m.group('value')
            cond = self.transform(value)
            cmd = f'__{key}__\n\t{cond}\n\t__end__\n'
            oslogger.debug(f'rewriting {value} to {cmd}')
            script = script.replace(m.group(), cmd)
        for m in RE_DRAW.finditer(script):
            _, args, kwargs = self.parse_cmd(m.group('cmd'))
            for key, val in kwargs.items():
                if key == 'show_if':
                    kwargs[key] = self.transform(val)
            cmd = self.create_cmd('draw', args, kwargs)
            oslogger.debug(f'rewriting {m.group("cmd")} to {cmd}')
            script = script.replace(m.group('cmd'), cmd)
        # We transform all Python f-strings into JavaScript template literals.
        # However, we ignore multiline _run and html variables, because those
        # can contain code that looks like f-strings but should not be
        # converted.
        start_pos = 0
        while True:
            ignore_spans = []
            for m in RE_IGNORE_RUN.finditer(script):
                ignore_spans.append(m.span())
            for m in RE_IGNORE_PREPARE.finditer(script):
                ignore_spans.append(m.span())
            for m in RE_IGNORE_SCRIPT.finditer(script):
                ignore_spans.append(m.span())
            for m in RE_FSTRING.finditer(script, start_pos):
                start, end = m.span()
                if any(ignore_start <= start and ignore_end >= end
                       for ignore_start, ignore_end in ignore_spans):
                    continue
                template_literal = f'${{{self.transform(m.group("expr"), escape=True)}}}'
                oslogger.debug(f'converting f-string {m.group()} to template '
                               f'literal {template_literal}')
                script = script[:start] + template_literal + script[end:]
                start_pos = end
                break
            else:
                break
        return script



"""This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public Lcense for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from bs4 import BeautifulSoup
from libopensesame.experiment import Experiment
from libopensesame.oslogging import oslogger
from pathlib import Path
import base64
import fileinspector
import hashlib
import io
import json
import os
import shutil
import sys
import tarfile
import tempfile
import time
import uuid
import zipfile
from . import version_info, sync
from .oswebwriter import OSWebWriter
from .oswebexceptions import VersionConflict, UnsupportedJZIP


FORMAT_IDENTIFIER = 'OSWeb JZIP Format v1'

# Paths towards assets that are bundled with the osweb extension code
src_folder = Path(os.path.dirname(__file__)) / 'src'
src_paths = {'js': src_folder / 'js',
             'css': src_folder / 'css',
             'html': src_folder / 'html',
             'img': src_folder / 'img'}


def supported_jzip(info_jas):
    """Takes a JSON dict in the format of info.jas and checks whether the
    info indicates that the associated JZIP can be converted to an Experiment.
    
    Parameters
    ----------
    info_jas: dict
    
    Returns
    -------
    bool
    """
    try:
        data = info_jas.get('data', {})
        if not isinstance(data, dict):
            oslogger.warning(
                f"Expected 'data' to be a dictionary in info_jas, got {data}")
            return False
        comments = data.get('comments', None)
        return comments == FORMAT_IDENTIFIER
    except Exception as e:
        oslogger.warning(f"An error occurred while checking info_jas: {e}")
        return False


def as_jatos_exp(exp):
    """Takes an experiment object (or a Path or str in which case the 
    experiment is opened from file), adds a UUID variable if it doesn't already
    exist, and returns the Experiment object.
    
    Parameters
    ----------
    exp : Experiment or str or Path
        Experiment
        
    Returns
    -------
    Experiment
    """
    if isinstance(exp, (str, Path)):
        oslogger.debug(f'reading experiment from file: {exp}')
        exp = Experiment(experiment_path=exp, string=exp)
    if not exp.var.has('jatos_uuid'):
        exp.var.jatos_uuid = unique_uuid()
        oslogger.debug(f'generating JATOS UUID: {exp.var.jatos_uuid}')
    return exp


def jzip_to_exp(jzip_path, factory=Experiment):
    """Converts a JZIP archive to an Experiment
    
    Parameters
    ----------
    jzip_path : str or Path
    factory : callable, optional
        A callable that takes a definition string and returns an Experiment
        object.
    
    Returns
    -------
    Experiment
    
    Raises
    ------
    UnsupportedJZIP
    """
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir_path = Path(tmp_dir)
        with zipfile.ZipFile(jzip_path, 'r') as zip_ref:
            zip_ref.extractall(tmp_dir)
        # Find the .jas file, which is the only file in the root with a .jas
        # extension
        info_jas_path = next(tmp_dir_path.glob('*.jas'))
        with info_jas_path.open('r') as f:
            info = json.load(f)
        if not supported_jzip(info):
            raise UnsupportedJZIP('Unsupported JZIP Study Archive')
        # Get the uuid and component_hash from info.jas
        uuid = info['data']['uuid']
        component_hash = info['data']['componentList'][0]['htmlFilePath'] \
            .replace('.html', '')
        # Build the paths to the necessary files
        osexp_path = tmp_dir_path / uuid / f'{component_hash}.orig.osexp'
        pool_path = tmp_dir_path / uuid / 'pool'
        version_info_path = tmp_dir_path / uuid / 'version_info.json'
        # Load the .orig.osexp file
        with osexp_path.open('r') as f:
            script = f.read()
        exp = factory(string=script)
        # Copy all files from the pool folder to the experiment's file pool.
        # If there are no files in the file pool, the pool path doesn't exist
        # in the jzip.
        dest_pool_path = Path(exp.pool.folder())
        if pool_path.exists():
            for filename in pool_path.iterdir():
                shutil.copy(filename, dest_pool_path)
        # Copy version_info.json to the experiment's file pool
        if version_info_path.exists():
            shutil.copy(version_info_path, dest_pool_path)
    return exp


def exp_to_jzip(exp, jzip_path=None, subject='0', fullscreen=False,
                welcome_text='', external_js=[], intro_click=True,
                full_background_color=False, jatos_info=None,
                ignore_conflicts=False):
    """Builds a jzip archive that can be imported into JATOS.

    Parameters
    ----------
    exp : Experiment or str or Path
        Experiment
    jzip_path : str or Path
        Path and filename of the resulting jzip file.  If no path is provided,
        a temporary file is created. The temporary needs to be explicitly
        removed.
    title : str, optional
        Title of the JATOS study, by default 'My OpenSesame experiment'.
    description : str, optional
        Description of the JATOS study, by default 'No description'.
    subject : str, optional
        Identifier of the JATOS subject, by default '0'.
    fullscreen : bool, optional
        Whether to run the JATOS study in fullscreen mode, by default False.
    welcome_text : str, optional
        Text to display on the JATOS study welcome screen, by default ''.
    external_js : List[str], optional
        List of URLs to external JavaScript libraties to include in the JATOS 
        study, by default [].
    intro_click: bool, optional
        Indicates whether the experiment should be preceded by a screen that
        the participant needs to click. This allows the experiment to be 
        executed in the context of a user action, which is necessary for
        certain actions.
    full_background_color: bool, optional
        Indicates whether full browser body should get the background color.
    jatos_info: JatosInfo or None, optional
        If provided, remote version information is checked and only modified
        assets are included in the archive.
    ignore_conflicts: bool, optional
        If True, conflicts result in a warning but do not trigger a 
        VersionConflict exception

    Returns
    -------
    Path
        The path to the jzip file
        
    Raises
    ------
    VersionConflict
    """
    exp = as_jatos_exp(exp)
    if jzip_path is None:
        jzip_path = Path(tempfile.NamedTemporaryFile(suffix='.jzip').name)
    else:
        jzip_path = Path(jzip_path)
    version_info_path = Path(exp.pool.folder()) / 'version_info.json'
    component_hash = unique_uuid()
    orig_script, osweb_script, pool_paths = _extract_script_and_pool_paths(exp)
    # We create a temporary directory that will contain all the files that will
    # go into the jzip archive.
    #
    # The folder structure is as follows:
    #
    #     info.jas                        # JATOS metadata
    #     {uuid}/
    #        {component_hash}.html        # HTML entry point
    #        {component_hash}.osexp       # Rewritten OSWeb script
    #        {component_hash}.orig.osexp  # Original OpenSesame script
    #        orig.osexp                   # As above but with constant name for
    #                                     # conflict detection
    #        version_info.json
    #        pool/
    #            {files from file pool}
    #        css/
    #            vendors-osweb*.css.map
    #            vendors-osweb*.css
    #            osweb*.css.map
    #            osweb*.css
    #            jatos.css
    #        js/
    #            vendors-osweb*.js.map
    #            vendors-osweb*.js
    #            osweb*.js.map
    #            osweb*.js
    #            jatos.js
    #        img/
    #            warning.png
    #            opensesame.png
    with tempfile.TemporaryDirectory(suffix='.jatos') as tmp_dir:
        tmp_dir = Path(tmp_dir)
        assets_dir = tmp_dir / exp.var.jatos_uuid
        assets_dir.mkdir()
        index_path = assets_dir / f'{component_hash}.html'
        jas_path = tmp_dir / 'info.jas'
        params = {'subject': subject,
                  'fullscreen': fullscreen,
                  'fullBackgroundColor': full_background_color,
                  'welcomeText': _safe_welcome_text(welcome_text),
                  'externalJS': external_js,
                  'introClick': intro_click}
        assets = _compose_html_and_get_assets(
            exp, index_path, 'jatos', params=params,
            component_hash=component_hash, pool_paths=pool_paths)
        info = {
            'version': '3',  # This refers to the JATOS version
            'data': {
                'title': exp.var.title,
                'description': exp.var.description,
                'groupStudy': False,
                'dirName': exp.var.jatos_uuid,
                'uuid': exp.var.jatos_uuid,
                'comments': FORMAT_IDENTIFIER,
                'jsonData': None,
                'componentList': [{
                    'title': 'OSWeb experiment',
                    'htmlFilePath': f'{component_hash}.html',
                    'reloadable': False,
                    'active': True,
                    'jsonData': json.dumps(params),
                }],
                'batchList': []
            }
        }
        if exp.var.get('jatos_end_redirect_url', ''):
            info['data']['endRedirectUrl'] = exp.var.jatos_end_redirect_url
        jas_path.write_text(json.dumps(info), 'utf-8')
        # Create a list of (src, jzip paths) where src can be a Path object
        # or a str that should be written to a file
        asset_list = [(index_path, f'{component_hash}.html'),
                      (osweb_script, f'{component_hash}.osexp'),
                      (orig_script, f'{component_hash}.orig.osexp'),
                      (orig_script, 'orig.osexp')]
        for pool_path in pool_paths:
            asset_list.append((pool_path, f'pool/{pool_path.name}'))
        for img in assets['img']:
            asset_list.append((img['src'], f'{img["dest"]}'))
        for js in assets['js']:
            asset_list.append((js['src'], f'{js["dest"]}'))
        for css in assets['css']:
            asset_list.append((css['src'], f'{css["dest"]}'))
        # Now that we have a complete list of files we can remove files that
        # don't need to be updated. To do so, we first update the current
        # version info. Then we try to download the previous version. If this
        # is not available, we upload everything. If it is available and there
        # is a conflict, then we raise an Exception. If it is available and
        # there is no conflict, then we remove all unmodified files
        if jatos_info is not None:
            asset_list.append((version_info_path, 'version_info.json'))
            current_info = version_info.update_version_info(exp, asset_list,
                                                            jatos_info)
            older_info = version_info.download_version_info(exp, jatos_info)
            if older_info is not None:
                vc = version_info.compare_version_info(jatos_info, older_info,
                                                       current_info)
                oslogger.debug(f'unmodified: {vc.unmodified}')
                oslogger.debug(f'modified: {vc.modified}')
                oslogger.debug(f'conflicting: {vc.conflicting}')
                oslogger.debug(f'added: {vc.added}')
                oslogger.debug(f'deleted: {vc.deleted}')
                if vc.conflicting:
                    file_list = '- ' + '\n- '.join(vc.conflicting)
                    msg = f'The current version of the experiment contains ' \
                          f'changes that conflict with the (remote) version ' \
                          f'of the experiment on JATOS.\n\nThis conflict ' \
                          f'results from the following files:\n\n{file_list}'
                    if ignore_conflicts:
                        oslogger.warning(msg)
                    else:
                        raise VersionConflict(msg)
            else:
                oslogger.debug(f'no older version info found')
                # TODO For now we don't strip the unmodified assets because
                # it's unclear how to do incremental uploads to JATOS
                # asset_list = [(src, tgt) for src, tgt in asset_list
                #               if tgt not in vc.unmodified]
        with zipfile.ZipFile(jzip_path, 'w') as fd:
            # info.jas is the only non-asset file
            fd.write(jas_path, 'info.jas')
            # Write all assets to jzip
            for src, tgt in asset_list:
                oslogger.debug(f'adding {tgt}')
                tgt = f'{exp.var.jatos_uuid}/{tgt}'
                if isinstance(src, Path):
                    fd.write(src, tgt)
                else:
                    fd.writestr(tgt, src)
    return jzip_path


def exp_to_html(exp, index_path=None, subject='0', logfile='osweb-data.json',
                fullscreen=False, welcome_text='', external_js=[],
                intro_click=True, full_background_color=False):
    """Builds an index.html that embeds everything and can be run in a browser.

    Parameters
    ----------
    exp : Experiment or str or Path
        Experiment
    index_path : str or Path or None, optional
        Path to the to-be-generated index.html file. If no path is provided,
        a temporary file is created. The temporary needs to be explicitly
        removed.
    subject : str, optional
        Identifier of the JATOS subject, by default '0'.
    logfile : str, optional
        The name of the logfile that is downloaded when the experiment is 
        finished.
    fullscreen : bool, optional
        Whether to run the JATOS study in fullscreen mode, by default False.
    welcome_text : str, optional
        Text to display on the JATOS study welcome screen, by default ''.
    external_js : List[str], optional
        List of URLs to external JavaScript libraties to include in the JATOS
        study, by default [].
    intro_click: bool, optional
        Indicates whether the experiment should be preceded by a screen that
        the participant needs to click. This allows the experiment to be 
        executed in the context of a user action, which is necessary for
        certain actions.
    full_background_color: bool, optional
        Indicates whether full browser body should get the background color.
        
    Returns
    -------
    Path
        The path to the html file
    """
    exp = as_jatos_exp(exp)
    _, script, pool_paths = _extract_script_and_pool_paths(exp)
    # We replace backslashes with slash forwards to avoid escaping issues with
    # first needing to escape the JSON and then needing to escape JavaScript
    # code
    params = {'subject': subject, 'logfile': logfile.replace('\\', '/'),
              'fullscreen': fullscreen,
              'fullBackgroundColor': full_background_color,
              'welcomeText': _safe_welcome_text(welcome_text),
              'externalJS': external_js, 'introClick': intro_click}
    if index_path is None:
        with tempfile.NamedTemporaryFile(suffix='.html', delete=False) as fd:
            index_path = fd.name
    index_path = Path(index_path)
    _compose_html_and_get_assets(exp, index_path, 'standalone', script=script,
                                 params=params, pool_paths=pool_paths)
    return index_path


# Private functions

def _get_os_assets(sub_dir):
    return [{'src': path, 'dest': f'{sub_dir}/{path.name}'}
            for path in src_paths[sub_dir].glob('*osweb*')]


def _compose_html_and_get_assets(exp, index_path, mode, script=None,
                                 params=None, component_hash=None,
                                 pool_paths=[]):
    """Generates an index.html and returns asset information.
    
    Parameters
    ----------
    exp: Experiment
    index_path: Path
        A path where the resulting index.html should be created
    mode: str
        'standalone' or 'jatos'
    script: str, optional
        The experiment script
    params: dict
        Experiment parameters, such as fullscreen
    component_hash: str, optional
    pool_paths: list, optional
        A list of paths that should be included in the file pool
        
    Returns
    -------
    dict
        A dict where keys are asset subdirectories ('js', 'css') and values are
        lists of files in the asset subdirectories.
    """
    assets = {'js': _get_os_assets('js'), 'css': _get_os_assets('css')}
    assets['img'] = [{'src': src_paths['img'] / 'opensesame.png',
                      'dest': 'img/opensesame.png'},
                     {'src': src_paths['img'] / 'osdoc.png',
                      'dest': 'img/osdoc.png'},
                     {'src': src_paths['img'] / 'warning.png',
                      'dest': 'img/warning.png'}]
    # Add potentially present platform js and css files.
    env_js = f'{mode}.js'
    env_js_src = src_paths['js'] / env_js
    if env_js_src.exists():
        assets['js'].append({'src': env_js_src, 'dest': (f'js/{env_js}')})
    assets['js'].append({'src': src_paths['js'] / '_shared.js',
                         'dest': ('js/_shared.js')})
    env_css = f'{mode}.css'
    env_css_src = src_paths[u'css'] / env_css
    if env_css_src.exists():
        assets['css'].append({'src': env_css_src, 'dest': (f'css/{env_css}')})
    tmpl = src_paths['html'] / f'{mode}.html'
    # The HTML file parsed as a DOM Tree
    dom = BeautifulSoup(tmpl.read_text('utf-8'), 'html.parser')
    if mode == 'standalone':
        _compose_for_standalone(script, dom, assets, params)
    elif mode == 'jatos':
        _compose_for_jatos(component_hash, exp.var.jatos_uuid, dom, assets,
                           params)
    # Add the file pool. This is a hidden div tag that contains elements for
    # each of the files in the pool. These are dynamically loaded by OSWeb.
    pool = dom.new_tag('div', id='filePool', style='display:none;')
    for path in pool_paths:
        # File inspector doesn't accept Path objects
        mime = fileinspector.determine_type(str(path))
        # This is a hack to make sure that CSV files are tagged with the
        # correct mimetype, instead of 'application/vnd.ms-excel', which
        # happens on Windows when Excel is installed.
        if str(path).lower().endswith('.csv'):
            mime = 'text/csv'
        if mime is None:
            oslogger.warning(f'unknown mimetype for {path}')
            continue
        if mode == 'standalone':
            src = f'data:{mime};base64,{_read_b64(path)}'
        else:
            src = Path('pool') / path.name
        id_ = path.name
        oslogger.debug(f'adding {path} to file pool')
        if mime.startswith('image/'):
            asset = dom.new_tag('img', src=src, id=id_)
        elif mime.startswith('text/'):
            asset = dom.new_tag('pre', id=id_)
            asset.string = path.read_text(encoding='utf-8')
        elif mime.startswith('video/'):
            src = dom.new_tag('source', src=src)
            asset = dom.new_tag('video', id=id_)
            asset.append(src)
        elif mime.startswith('audio/'):
            src = dom.new_tag('source', src=src)
            asset = dom.new_tag('audio', id=id_)
            asset.append(src)
        else:
            oslogger.warning(f'unsupported {mime} mimetype for {path}')
            continue
        pool.append(asset)
    dom.body.append(pool)
    # Write HTML file
    html = dom.prettify()
    index_path.write_text(html, 'utf-8')
    return assets


def _compose_for_standalone(script, dom, assets, params=None):
    """Builds on top of the base HTML template to create a structure that is 
    appropriate for a standalone HTML file.
    
    Parameters
    ----------
    script: str
    dom: Tag
        A BeautifulSoup Tag corresponding to the full HTML file
    assets: dict
    params: dict
        Experiment parameters, such as fullscreen
    """
    logo_tag = dom.new_tag(u'script', type=u'text/javascript')
    logo_path = src_paths['img'] / 'opensesame.png'
    logo_tag.append(
        f'const logoSrc = "data:image/png;base64,{_read_b64(logo_path)}"')
    dom.head.append(logo_tag)
    if params:
        # Embed links to external JavaScript pacakages
        for url in params['externalJS']:
            params_tag = dom.new_tag(u'script', type=u"text/javascript",
                                     src=url.strip())
            dom.head.append(params_tag)
        params_tag = dom.new_tag(u'script', type=u"text/javascript")
        params_tag.append(
            f'const params = JSON.parse(\'{json.dumps(params)}\')\n')
        dom.head.append(params_tag)
    script_tag = dom.new_tag('script', type='text/javascript')
    for js_file in assets['js']:
        if js_file['src'].suffix == '.js':
            script_tag.append(js_file['src'].read_text('utf-8') + '\n')
    dom.head.append(script_tag)
    css_tag = dom.new_tag(u'style')
    for css_file in assets['css']:
        if css_file['src'].suffix == '.css':
            css_tag.append(css_file['src'].read_text('utf-8') + '\n')
    dom.head.append(css_tag)
    # Add experiment as base64 encoded string
    exp_tag = dom.new_tag(
        'embed', id='osexp_src',
        src=f'data:text/plain;base64,{_to_b64(script)}',
        style='display:none')
    dom.body.append(exp_tag)


def _compose_for_jatos(component_hash, uuid, dom, assets, params):
    """Builds on top of the base HTML template to create a structure that is 
    appropriate for integration in JATOS.
    
    Parameters
    ----------
    component_hash: str
    uuid: str
    dom: Tag
        A BeautifulSoup Tag corresponding to the full HTML file
    assets: dict
    params: dict
        Experiment parameters, such as fullscreen
    """
    script_tag = dom.new_tag('script', id="parameters", type="text/javascript")
    if params:
        # Embed links to external JavaScript pacakages
        for url in params['externalJS']:
            params_tag = dom.new_tag(u'script', type=u"text/javascript",
                                     src=url)
            dom.head.append(params_tag)        
        # Create a script node that exposes the experiment's parameters
        script_tag.append(
            f'const params = JSON.parse(\'{json.dumps(params)}\')')
    # Get the OpenSesame experiment file name, and add a JS variable to its
    # location
    script_tag.append(f'const osexpFile = "{component_hash}.osexp"')
    script_tag.append(f'const jatosAssetsFolder = "{uuid}"')
    script_tag.append(f'const jatosComponentHash = "{component_hash}"')
    dom.head.append(script_tag)
    # Add script nodes referencing the sources of all other required javascript
    # files.
    for js_file in assets['js']:
        if js_file['src'].suffix == '.js':
            script_tag = dom.new_tag('script', src=js_file['dest'],
                                     type="text/javascript")
            dom.head.append(script_tag)
    for css_file in assets['css']:
        if css_file['src'].suffix == '.css':
            styleTag = dom.new_tag('link', href=css_file['dest'],
                                   type="text/css", rel="stylesheet",
                                   media="all")
            dom.head.append(styleTag)


def _read_b64(path):
    return base64.b64encode(path.read_bytes()).decode()
    

def _to_b64(s):
    return base64.b64encode(s.encode('utf-8')).decode()


def unique_uuid():
    return str(uuid.UUID(hashlib.md5(str(time.time()).encode()).hexdigest()))


def _safe_welcome_text(s):
    return s.replace('\n', '<br />'). \
        replace('"', '&#34;').replace("'", '&#39;')


def _extract_script_and_pool_paths(exp):
    """Extracts the scripts and pool paths from the experiment. There are
    two scripts. The first is the original script. The second is the script as
    it will be used by OSWeb, which is partly rewritten. If the experiment is
    not configured to use OSWeb, the second script is a placeholder that 
    informs the user that the experiment does not run in a browser.
    """
    if exp.var.canvas_backend != 'osweb':
        oslogger.debug('using placeholder experiment')
        osweb_exp = as_jatos_exp(src_folder / 'osexp/placeholder.osexp')
    else:
        osweb_exp = exp
    with tempfile.NamedTemporaryFile(suffix='.osexp', delete=False) as fd:
        pass
    script = OSWebWriter(osweb_exp, fd.name).script
    try:
        os.remove(fd.name)
    except Exception as e:
        oslogger.warning(f'failed to remove temporary file: {fd.name}')
    pool_paths = [Path(exp.pool[path]) for path in exp.pool]
    return exp.to_string(), script, pool_paths



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.runners.base_runner import BaseRunner


class OSWebRunner(BaseRunner):
    """Runs an experiment in the traditional way, in the same process."""
    
    def execute(self):
        return_values = self.main_window.extension_manager.fire(
            'osweb_run', fullscreen=self._fullscreen,
            subject_nr=self._subject_nr, logfile=self._logfile)
        return return_values.get('Oswebext', None)
    
    def init_experiment(self, quick=False, fullscreen=False):
        # Get and set the subject number
        subject_nr = self.get_subject_nr(quick=quick)
        if subject_nr is None:
            return False
        # Get and set the logfile
        logfile = self.get_logfile(quick=quick, subject_nr=subject_nr,
                                   file_extension='.json')
        if logfile is None:
            return False
        self._subject_nr = subject_nr
        self._logfile = logfile
        self._fullscreen = fullscreen
        return True

    def workspace_globals(self):
        return {}



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import webbrowser
import os
from pathlib import Path
from libopensesame.exceptions import UserAborted
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
from libqtopensesame.misc.config import cfg
from libqtopensesame.items.experiment import Experiment
from . import osweb_linter as linter
from osweb.oswebexceptions import PythonToJavaScriptError, \
    OSWebCompatibilityError, ListRemoteError, UnsupportedJZIP, \
    JZIPDownloadError
from openexp import backend
from qtpy.QtWidgets import QFileDialog, QMessageBox, QApplication, \
    QProgressDialog
from qtpy.QtCore import Qt
_ = translation_context('oswebext', category='extension')


def decorate_backend_info(fnc):
    """Wraps around the backend_info() function to add the OSWeb backend."""
    def inner():
        backend_info = fnc()
        backend_info['osweb'] = {'description': 'In a browser with OSWeb',
                                 'canvas': 'osweb',
                                 'keyboard': 'osweb',
                                 'mouse': 'osweb',
                                 'sampler': 'osweb',
                                 'color': 'osweb',
                                 'clock': 'osweb',
                                 'log': 'osweb',
                                 'icon': 'applications-internet',
                                 'settings': False}
        return backend_info
    return inner


class Oswebext(BaseExtension):
    
    preferences_ui = 'extensions.oswebext.preferences'

    def event_startup(self):
        cfg.oswebext_jatos_ignore_conflicts = False
        self._control_panel = None
        self._experiment_name = None
        self._experiment_path = None
        backend.backend_info = decorate_backend_info(backend.backend_info)
        file_menu = self.get_submenu('file')
        if cfg.oswebext_jatos_url == 'https://jatos.mindprobe.eu':
            open_title = _('Open from MindProbe')
            publish_title = _('Save and publish to MindProbe')
        else:
            open_title = _('Open from JATOS')
            publish_title = _('Save and publish to JATOS')
        action_open_jatos = self.qaction('document-open',
                                         open_title,
                                         self.event_osweb_open_jatos)
        action_publish_jatos = self.qaction('document-save',
                                            publish_title,
                                            self.event_osweb_publish_jatos)
        self.add_action(file_menu, action_open_jatos, -4, False, False)
        self.add_action(file_menu, action_publish_jatos, -4, False, True)
        
    def provide_runner(self):
        if self.experiment.var.canvas_backend == 'osweb':
            from .osweb_runner import OSWebRunner
            return OSWebRunner
        
    def provide_item_supported(self, experiment, item_type):
        if self.experiment.var.canvas_backend != 'osweb':
            return None
        return linter.item_type_supported(item_type)
        
    def _suggested_path(self, suffix):
        folder = os.path.dirname(cfg.file_dialog_path)
        name = self.experiment.syntax.sanitize(
            self.experiment.var.title, strict=True, allow_vars=False) + suffix
        return os.path.join(folder, name)
        
    def event_osweb_export_html(self):
        from osweb import convert
        path = QFileDialog.getSaveFileName(
            self.main_window, _('Save as…'),
            directory=self._suggested_path('.html'),
            filter='HMTL files (.html)')
        # In PyQt5, the QFileDialog.getOpenFileName returns a tuple instead of
        # a string, of which the first position contains the path.
        if isinstance(path, tuple):
            path = path[0]
        if not path:
            return
        if not path.lower().endswith('.html'):
            path += '.html'
        try:
            if not cfg.oswebext_bypass_linter:
                errors = linter.check_compatibility(
                    self.experiment, fullscreen=cfg.oswebext_fullscreen)
                if errors:
                    raise OSWebCompatibilityError(errors)
            convert.exp_to_html(
                self.experiment,
                index_path=path,
                subject=str(cfg.oswebext_subject_nrs),
                logfile='osweb-data.json',
                fullscreen=cfg.oswebext_fullscreen,
                welcome_text=cfg.oswebext_welcome_text,
                external_js=self._external_js(),
                intro_click=cfg.oswebext_intro_click)
        except (OSWebCompatibilityError, PythonToJavaScriptError) as e:
            self.report_exception(e)
            return
        self._refresh_control_panel()

    def event_osweb_export_jzip(self):
        from osweb import convert
        path = QFileDialog.getSaveFileName(
            self.main_window, _('Save as…'),
            directory=self._suggested_path('.jzip'),
            filter='JATOS Study Archive (*.jzip)')
        # In PyQt5, the QFileDialog.getOpenFileName returns a tuple instead of
        # a string, of which the first position contains the path.
        if isinstance(path, tuple):
            path = path[0]
        if not path:
            return
        if not path.lower().endswith('.jzip'):
            path += '.jzip'
        convert.exp_to_jzip(
            self.experiment,
            jzip_path=path,
            subject=str(cfg.oswebext_subject_nrs),
            fullscreen=cfg.oswebext_fullscreen,
            full_background_color=cfg.oswebext_full_background_color,
            welcome_text=cfg.oswebext_welcome_text,
            external_js=self._external_js(),
            intro_click=cfg.oswebext_intro_click)
        self._refresh_control_panel()
        
    def event_osweb_import_jzip(self):
        from osweb import convert
        path = QFileDialog.getOpenFileName(
            self.main_window, _('Open'),
            directory=os.path.dirname(cfg.file_dialog_path),
            filter='JATOS Study Archive (*.jzip)')
        # In PyQt5, the QFileDialog.getOpenFileName returns a tuple instead of
        # a string, of which the first position contains the path.
        if isinstance(path, tuple):
            path = path[0]
        if not path:
            return
        self._open_jzip(path)

    def event_osweb_open_jatos(self):
        from osweb import sync
        if not self._jatos_configured():
            return
        self.main_window.set_busy(True)
        try:
            remote_exp_list = sync.list_remote(self._jatos_info())
        except ListRemoteError as e:
            self.report_exception(e)
            return
        finally:
            self.main_window.set_busy(False)
        haystack = []
        for remote_exp in remote_exp_list:
            haystack.append(
                (remote_exp['title'], remote_exp,
                 self._select_remote_experiment))
        self.extension_manager.fire(
            'quick_select',
            haystack=haystack,
            placeholder_text=_('Browse') + f' {cfg.oswebext_jatos_url} …')
        
    def _select_remote_experiment(self, remote_exp):
        from osweb import sync, convert
        self.main_window.set_busy(True)
        self._progress_dialog = QProgressDialog(
            _('Downloading …'), _('Hide'), 0, 100, self.main_window)
        self._progress_dialog.setWindowModality(Qt.WindowModal)
        self._progress_dialog.show()
        self._progress_dialog.setWindowTitle(_('Please wait …'))
        try:
            jzip_path = sync.download_jzip(remote_exp['uuid'],
                                           self._jatos_info(),
                                           callback=self._progress_callback)
        except JZIPDownloadError as e:
            self.report_exception(e)
        else:
            self._open_jzip(jzip_path)
        finally:
            self.main_window.set_busy(False)
            
    def _progress_callback(self, transferred, total):
        progress_percentage = int((transferred / total) * 100)
        self._progress_dialog.setValue(progress_percentage)
        QApplication.processEvents()
    
    def event_osweb_publish_jatos(self):
        from osweb import sync, convert
        if not self._jatos_configured():
            return
        self.main_window.set_busy(True)
        self.main_window.save_file()
        self._progress_dialog = QProgressDialog(
            _('Uploading …'), _('Hide'), 0, 100, self.main_window)
        self._progress_dialog.setWindowModality(Qt.WindowModal)
        self._progress_dialog.show()
        self._progress_dialog.setWindowTitle(_('Please wait …'))
        try:
            sync.upload(
                self.experiment,
                self._jatos_info(),
                subject=str(cfg.oswebext_subject_nrs),
                fullscreen=cfg.oswebext_fullscreen,
                full_background_color=cfg.oswebext_full_background_color,
                welcome_text=cfg.oswebext_welcome_text,
                external_js=self._external_js(),
                intro_click=cfg.oswebext_intro_click,
                ignore_conflicts=cfg.oswebext_jatos_ignore_conflicts,
                callback=self._progress_callback)
        except Exception as e:
            self.report_exception(e)
        else:
            self.extension_manager.fire(
                'notify',
                message=_('Experiment has been published to JATOS'),
                always_show=True)
        finally:
            self.main_window.set_busy(False)
        # Automatically reset this option to avoid accidental force uploads in
        # the future
        cfg.oswebext_jatos_ignore_conflicts = False
        self.extension_manager.fire('setting_changed',
                                    setting='oswebext_jatos_ignore_conflicts',
                                    value=False)
        self._refresh_control_panel()
    
    def event_osweb_convert_results(self):
        from osweb import results
        from datamatrix import io

        results_path = QFileDialog.getOpenFileName(
            self.main_window, _('Select OSWeb results file…'),
            filter='OSWeb results (*.*)')
        if isinstance(results_path, tuple):
            results_path = results_path[0]
        if not results_path:
            return
        results_path = Path(results_path)
        self.main_window.set_busy(True)
        try:
            dm = results.parse_results(
                results_path, include_context=cfg.oswebext_include_context)
        except Exception as e:
            self.report_exception(e)
            return
        finally:
            self.main_window.set_busy(False)
        export_path = QFileDialog.getSaveFileName(
            self.main_window, _('Save as…'),
            filter='Excel (*.xlsx);;CSV (*.csv)')
        if isinstance(export_path, tuple):
            export_path = export_path[0]
        if not export_path:
            return
        export_path = Path(export_path)
        if export_path.suffix.lower() not in ('.csv', '.xlsx'):
            export_path = export_path.with_suffix('.xlsx')
        if export_path.suffix.lower().endswith('.xlsx'):
            io.writexlsx(dm, export_path)
        else:
            io.writetxt(dm, export_path)
        
    def event_osweb_run(self, fullscreen=False, subject_nr=0,
                        logfile='quickrun.csv'):
        from osweb import convert
        if not cfg.oswebext_bypass_linter:
            errors = linter.check_compatibility(
                self.experiment, fullscreen=cfg.oswebext_fullscreen)
            if errors:
                return OSWebCompatibilityError(errors)
        # Replace the csv extension by a json extension
        if logfile.lower().endswith('.csv'):
            logfile = logfile[:-4] + '.json'
        self.main_window.get_ready()
        try:
            index_path = convert.exp_to_html(
                self.experiment,
                subject=str(subject_nr),
                logfile=logfile,
                fullscreen=fullscreen,
                full_background_color=cfg.oswebext_full_background_color,
                welcome_text=cfg.oswebext_welcome_text,
                external_js=self._external_js(),
                intro_click=cfg.oswebext_intro_click)
        except (OSWebCompatibilityError, PythonToJavaScriptError) as e:
            return e
        else:
            webbrowser.open('file://{}'.format(index_path))
            # The UserAborted exception will result in a notification, rather
            # than # in a full-screen tab saying that the experiment has
            # finished.
            return UserAborted(_('Experiment started in external browser'))
        finally:
            self._refresh_control_panel()
            
    def event_prepare_change_experiment(self):
        self._experiment_name = self.experiment.var.title
    
    def event_change_experiment(self):
        """When the experiment title changes, we offer to reset the JATOS
        UUID.
        """
        if not self.experiment.var.has('jatos_uuid'):
            return
        if self._experiment_name == self.experiment.var.title:
            return
        resp = QMessageBox.question(
            self.main_window,
            _('Unlink from JATOS'),
            _('You have changed the name of the experiment. Do you also want to unlink the experiment from JATOS (by resetting the UUID) so that you can create a new remote experiment?'),
            QMessageBox.Yes,
            QMessageBox.No
        )
        if resp == QMessageBox.No:
            return
        # Delete the JATOS uuid and close all other tabs so that we don't have
        # to update the preferences or control panel
        del self.experiment.var.jatos_uuid
        self.tabwidget.close_other()
        
    def event_save_experiment(self, path=None):
        """If an experiment is saved under a different name, we offer to reset
        the JATOS UUID.
        """
        if self.experiment.var.has('jatos_uuid') and \
                self._experiment_path is not None and \
                self._experiment_path != path:
            resp = QMessageBox.question(
                self.main_window,
                _('Unlink from JATOS'),
                _('You have saved the experiment under a different file name. Do you also want to unlink the experiment from JATOS (by resetting the UUID) so that you can create a new remote experiment?'),
                QMessageBox.Yes,
                QMessageBox.No
            )
            if resp == QMessageBox.Yes:
                # Delete the JATOS uuid and close all other tabs so that we don't have
                # to update the preferences or control panel
                del self.experiment.var.jatos_uuid
                self.tabwidget.close_other()
        self._experiment_path = path
    
    def event_open_experiment(self, path=None):
        self._experiment_path = path
        
    def activate(self):
        self._show_controls()
        
    def settings_widget(self):
        widget = super().settings_widget()
        widget.ui.button_clear_jatos_uuid.clicked.connect(
            lambda: self._clear_jatos_uuid(widget))
        widget.ui.edit_end_redirect_url.textEdited.connect(
            self._update_end_redirect_url)
        return widget
    
    def _update_end_redirect_url(self, text):
        self.experiment.var.jatos_end_redirect_url = text
        
    def _clear_jatos_uuid(self, widget):
        del self.experiment.var.jatos_uuid
        widget.ui.edit_uuid.setText('')

    def _show_controls(self, jatos_configured=True):
        from .osweb_control_panel import OSWebControlPanel
        if self._control_panel is None:
            self._control_panel = OSWebControlPanel(self.main_window,
                                                    self.settings_widget())
        self.tabwidget.add(self._control_panel, self.icon(), self.label())
        self._control_panel.set_jatos_configured(jatos_configured)
        
    def _external_js(self):
        print('external js =', cfg.oswebext_external_js)
        return [url.strip() for url in cfg.oswebext_external_js.splitlines()]

    def _jatos_configured(self):
        if isinstance(cfg.oswebext_jatos_url, str) and \
                cfg.oswebext_jatos_url.startswith('http') and \
                isinstance(cfg.oswebext_jatos_api_token, str) and \
                cfg.oswebext_jatos_api_token.startswith('jap_'):
            return True
        self._show_controls(jatos_configured=False)
        return False

    def _jatos_info(self):
        from osweb import sync
        return sync.JatosInfo(cfg.oswebext_jatos_url,
                              cfg.oswebext_jatos_api_token)

    def _refresh_control_panel(self):
        if self._control_panel is None:
            return
        self._control_panel.refresh()

    def _open_jzip(self, jzip_path):
        from osweb import convert
        try:
            exp = convert.jzip_to_exp(
                jzip_path,
                factory=lambda string: Experiment(self.main_window,
                                                  string=string))
        except UnsupportedJZIP as e:
            self.report_exception(e)
        else:
            self.main_window.ui.tabwidget.close_all(avoid_empty=False)
            self.main_window.open_experiment(exp)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import itertools as it
from libqtopensesame.misc.translate import translation_context
_ = translation_context('oswebext', category='extension')


# These items are supported but don't have a specific linter function
SUPPORTED_ITEMS = [
    'sequence',
    'touch_response',
    'notepad',
    'reset_feedback',
    'repeat_cycle',
    'advanced_delay',
    'inline_javascript',
    'inline_html',
    'feedback',
    'form_text_display',
    'form_multiple_choice',
    'form_consent',
    'logger',
    'sketchpad'
]

STRUCTURE_CHECK_ITEMS = [
    'repeat_cycle',
    'reset_feedback',
    'loop',
    'sequence'
]

MAX_SIZE = 10 * 1024 ** 2  # 10 MB


def item_type_supported(item_type):
    """Checks whether an item_type is supported. Items are supported if they
    have a custom linter function or if they are in the SUPPORTED_ITEMS list.
    """
    try:
        eval('check_item_{}'.format(item_type))
    except NameError:
        return item_type in SUPPORTED_ITEMS
    return True


def check_compatibility(exp, fullscreen):
    """Checks the compatibility of an experiment and returns all warnings as a
    list of text strings. An empty list indicates that the experiment is
    compatible.
    """
    errors = check_supported_items(exp, fullscreen) \
        + check_structure(exp, fullscreen) \
        + check_filesize(exp)
    if not errors:
        return ''
    return '- ' + '\n - '.join(errors)


def check_item(item, fullscreen):
    """Checks whether an item is compatible and returns all warnings as a list
    of text strings. An empty list indicates that the experiment is
    compatible. If an item specific linter function is available (with the 
    name check_item_[item type]) then this is used. Otherwise, a warning is
    giving when the item type is not in the list of supported items.
    """
    try:
        linter_fnc = eval('check_item_{}'.format(item.item_type))
    except NameError:
        if item.item_type not in SUPPORTED_ITEMS:
            return ['Item {} is not supported'.format(item.item_type)]
        return []
    return linter_fnc(item, fullscreen)


def check_supported_items(exp, fullscreen):
    """Checks all items and returns a list of warnings."""
    return list(it.chain(
        *[check_item(item, fullscreen) for item in exp.items.values()]
    ))


def check_structure(exp, fullscreen):
    """Checks the structure of the experiment for a specific issue with OSWeb
    in which for certain items the prepare and run phases need to be called
    in alternation, i.e. it is not allowed to call the prepare phase (or the
    run phase) twice in a row.
    """
    def set_state(item_name, state):
        
        if item_name not in exp.items:
            return
        item = exp.items[item_name]
        if (
            states.get(item_name, None) == state and
            item.item_type in STRUCTURE_CHECK_ITEMS
        ):
            errors.append(
                'The {} phase for item {} is called multiple times in a row'.format(
                    state,
                    item_name
                )
            )
        states[item_name] = state
        if item.item_type == 'sequence':
            for sub_item_name, _, _ in item.items:
                set_state(sub_item_name, state)
        elif item.item_type == 'loop':
            set_state(item._item, 'prepare')
            set_state(item._item, 'run')

    errors = []
    states = {}
    set_state(exp.var.start, 'prepare')
    set_state(exp.var.start, 'run')
    return errors


def item_warning(item, msg):
    return '{} ({}): {}'.format(item.name, item.item_type, msg)


def check_item_loop(item, fullscreen):
    w = []
    if item.var.continuous != 'no':
        w.append(item_warning(item, 'Resume after break not supported'))
    if item.var.break_if_on_first != 'yes':
        w.append(
            item_warning(
                item,
                'Disabling evaluate on first cycle not supported'
            )
        )
    if item._constraints:
        w.append(item_warning(item, 'Constraints not supported'))
    return w

    
def check_item_mouse_response(item, fullscreen):
    w = []
    if item.var.event_type == 'mouserelease':
        w.append(item_warning(item, 'Mouse release not supported'))
    return w


def check_item_keyboard_response(item, fullscreen):
    w = []
    if item.var.event_type == 'keyrelease':
        w.append(item_warning(item, 'Key release not supported'))
    return w


def check_item_sampler(item, fullscreen):
    w = []
    if item.var.stop_after != 0:
        w.append(item_warning(item, 'Stop after not supported'))
    return w


def check_item_form_text_input(item, fullscreen):
    if not fullscreen:
        return []
    return [item_warning(
        item,
        'Some browsers disable text input in fullscreen for security reasons'
    )]


def check_filesize(exp):
    try:
        size = exp.pool.size()
    except Exception:
        return []
    if size < MAX_SIZE:
        return []
    return [f'The experiment is {size / 1024 ** 2:.0f} MB, which exceeds the '
            f'recommended maximum size']



# coding=utf-8
from libopensesame.py3compat import *
from libopensesame.item import Item
from libopensesame.oslogging import oslogger


class InlineHtml(Item):

    def reset(self):
        
        self.var.html = '''<input type='text' name='text_response'>
<input type='submit' value='ok'>'''

    def run(self):
        
        oslogger.warning(
            'inline_html items require OSWeb and do nothing on the desktop'
        )



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import json
from datamatrix import DataMatrix, operations as ops
from pathlib import Path
import tempfile
import zipfile
from datamatrix.py3compat import *


STEPS = 100


def parse_results(path, include_context=False):
    """Converts a results file, detects its type, and returns it as a
    DataMatrix. If the context is included, columns are created for each of the
    context variables.

    Parameters
    ----------
    path: Path
    include_context: bool

    Returns
    -------
    DataMatrix
    """
    path = Path(path)
    if path.suffix.lower() in ('.zip', '.jrzip'):
        return parse_jrzip(path, include_context)
    if path.suffix.lower() == '.json':
        return parse_json(path)
    return parse_jatos_results(path, include_context)
    
    
def parse_json(path):
    """Converts a json results file, as stored by standalone OSWeb, and returns
    it as a DataMatrix. If the context is included, columns are created for
    each of the context variables.

    Parameters
    ----------
    path: Path
    include_context: bool

    Returns
    -------
    DataMatrix
    """
    data = json.loads(path.read_text())
    if not isinstance(data, list):
        raise TypeError('Invalid JSON results file')
    dm = DataMatrix(length=len(data))
    for dm_row, data_row in zip(dm, data):
        for key, val in data_row.items():
            dm_row[key] = val
    return dm


def parse_jrzip(path, include_context):
    """Converts a JATOS results archive and returns it as a DataMatrix. If the
    context is included, columns are created for each of the context variables.

    Parameters
    ----------
    path: Path
    include_context: bool

    Returns
    -------
    DataMatrix
    """
    dms = []
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = Path(temp_dir)
        with zipfile.ZipFile(path, 'r') as zip_ref:
            zip_ref.extractall(temp_dir)
            # The archive contains separate txt files that are similar to one
            # line from the old txt data files.
            for results_file in temp_dir.glob('**/*.txt'):
                dm = parse_jatos_results(results_file, include_context)
                # We extract the jatosResultId from the file name, which has
                # the format study_result_[result_id]
                dm.jatosStudyResultId = results_file.parts[-3].split('_')[-1]
                dms.append(dm)
    return ops.stack(dms)


def parse_jatos_results(path, include_context):
    """Converts a JATOS data-only results file and returns it as a DataMatrix.
    If the context is included, columns are created for each of the context
    variables.
    
    Parameters
    ----------
    path: Path
    include_context: bool
    
    Returns
    -------
    DataMatrix
    """
    if hasattr(json.decoder, 'JSONDecodeError'):
        jsonerror = json.decoder.JSONDecodeError
    else:
        jsonerror = ValueError
    dm = DataMatrix(length=STEPS)
    invalid_lines = 0
    incomplete_lines = 0
    total_lines = 0
    row = 0
    with path.open(encoding='utf-8') as fd:
        for line in fd:
            # Strip the lines so that they are valid json
            line = line.strip()
            if line.startswith('[{'):
                line = line[1:]
            if line.endswith('},') or line.endswith('}]'):
                line = line[:-1]
            total_lines += 1
            # Ignore empty lines
            if not line:
                continue
            try:
                d = json.loads(line)
            except jsonerror:
                invalid_lines += 1
                continue
            else:
                if not d:
                    continue
            # Complete data is stored as a single-json object with a data and
            # context field. We're only interested in the data field here,
            # which in turn is a list of dicts where each dict corresponds to
            # a logger call. This data is stored when an experiment is
            # successfully finished.
            if len(d) == 2 and 'data' in d and 'context' in d:
                trials = d['data']
                context = d['context']
            # Incomplete data is stored as a single json line corresponding to
            # a single logger call. This kind of data should only happen when
            # the experiment is not finished, so that data is only partially
            # logged, one trial at a time.
            else:
                trials = [d]
                context = {}
                incomplete_lines += 1
            first_row = row
            for trial in trials:
                if row >= len(dm):
                    dm.length += STEPS
                for key, val in trial.items():
                    if key not in dm:
                        dm[key] = u''
                    dm[key][row] = safe_decode(val)
                row += 1
            if include_context:
                for key, value in _flatten_dict(context).items():
                    if key not in dm:
                        dm[key] = u''
                    dm[key][first_row:row] = safe_decode(value)
    dm.length = row
    if invalid_lines:
        warn('Failed to parse {} of {} lines'.format(invalid_lines,
                                                     total_lines))
    if incomplete_lines:
        warn('Incomplete (unfinished) data in {} of {} lines'.format(
            incomplete_lines, total_lines))
    return dm


def _flatten_dict(d):
    
    flat_dict = {}
    for key, value in d.items():
        if isinstance(value, dict):
            for subkey, subvalue in _flatten_dict(value).items():
                flat_dict['{}_{}'.format(key, subkey)] = subvalue
        else:
            flat_dict[key] = value
    return flat_dict


if __name__ == '__main__':

    import sys
    dm = parse_jatos_results(sys.argv[-1])
    print(dm)
    print('{} trials'.format(len(dm)))



'''Embeds custom HTML'''

category = 'OSWeb'
controls = [{
    'type': 'editor',
    'var': 'html',
    'label': 'HTML editor',
    'syntax': True,
    'language': 'html',
    'name': 'editor_widget',
    'tooltip': 'An HTML editor widget'
}]
def supports(exp):
    return exp.var.canvas_backend == 'osweb'



# coding=utf-8
from libopensesame.py3compat import *
import js2py
from js2py.translators import translate_js
import esprima
from esprima.nodes import FunctionDeclaration, VariableDeclaration, \
    ExpressionStatement, AssignmentExpression, StaticMemberExpression, \
    BlockStatement
from libopensesame.inline_script import InlineScript
from libqtopensesame.items.inline_script import InlineScript as QtInlineScript
from libqtopensesame.items.qtplugin import QtPlugin
from pyqode.core.widgets import SplittableCodeEditTabWidget
from libqtopensesame.misc.translate import translation_context
from libopensesame.oslogging import oslogger
from . import javascript_workspace_api as api
_ = translation_context(u'inline_javascript', category=u'plugin')


class JavaScriptWorkspace(js2py.EvalJs):

    def __init__(self, experiment):

        self.__dict__[u'experiment'] = experiment
        self.__dict__[u'_globals'] = {}
        self.__dict__[u'_persistent'] = {}
        self.init_globals()
        js2py.EvalJs.__init__(self, self._globals)

    def init_globals(self):

        """
        desc:
            Initialize the global workspace.
        """
        # By setting the __name__ global, the workspace will operate as a
        # module, so that e.g. import statements don't suffer from locality.
        self._globals.update({
            u'__name__': u'javascript_workspace',
            u'vars': self.experiment.var,
            u'persistent': self._persistent
        })
        for name, obj in api.__dict__.items():
            if name.startswith('_'):
                continue
            self._globals[name] = obj

    def _compile(self, js=None, use_compilation_plan=False):

        code = translate_js(
            js,
            '',
            use_compilation_plan=use_compilation_plan
        )
        return compile(
            code,
            '<EvalJS snippet>',
            'exec'
        )

    def _exec(self, compiled):

        exec(compiled, self._context)


class InlineJavascript(InlineScript):

    def reset(self):

        self.var._prepare = u''
        self.var._run = u''

    @property
    def workspace(self):

        if not hasattr(self.experiment, u'javascript_workspace'):
            self.experiment.javascript_workspace = JavaScriptWorkspace(
                self.experiment)
        return self.experiment.javascript_workspace


class QtInlineJavascript(QtInlineScript):

    language = u'JavaScript'  # For OpenSesame 3.2
    ext = u'.js'  # For OpenSesame 3.3
    mime_type = u'application/javascript'  # For OpenSesame 3.3

    def __init__(self, name, experiment, string=None):

        # This requires pyqode_extras, which is not available in opensesamerun
        # and therefore we import it only here.
        from .javascript_code_edit import JavaScriptCodeEdit

        self._var_cache = None
        if self.mime_type not in SplittableCodeEditTabWidget.editors:
            SplittableCodeEditTabWidget.register_code_edit(JavaScriptCodeEdit)
        InlineScript.__init__(self, name, experiment, string)
        QtPlugin.__init__(self, plugin_file=__file__)

    def item_icon(self):

        return self.qicon

    @staticmethod
    def _extract_assignments(script):
    
        def inner(body, only_globals=False):
            if hasattr(body, 'body') and body.body is not None:
                return inner(body.body,
                    only_globals=only_globals or 
                        isinstance(body, FunctionDeclaration))
            if not isinstance(body, list):
                body = [body]
            assignments = []
            for element in body:
                if isinstance(element, VariableDeclaration) and \
                        not only_globals:
                    # Only var results in assignments
                    if element.kind == 'var':
                        for declaration in element.declarations:
                            assignments.append(declaration.id.name)
                elif isinstance(element, ExpressionStatement) \
                        and isinstance(element.expression,
                                       AssignmentExpression) \
                        and isinstance(element.expression.left,
                                       StaticMemberExpression):
                    if element.expression.left.object.name \
                            in ['vars','window']:
                        assignments.append(
                            element.expression.left.property.name)
                elif isinstance(element, list):
                    assignments += inner(body, only_globals=only_globals)
                elif hasattr(element, 'body') and element.body is not None:
                    assignments += inner(element.body,
                        only_globals=only_globals or 
                            isinstance(element, FunctionDeclaration))
            return assignments
        
        try:
            return inner(esprima.parse(script).body)
        except Exception as e:
            oslogger.debug(f'failed to extract assignments: {e}')
            return []
    



#-*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import sys
from libopensesame.oslogging import oslogger
from pyqode.python.backend import server
from libqtopensesame.pyqode_extras.widgets import FallbackCodeEdit


class JavaScriptCodeEdit(FallbackCodeEdit):

    mimetypes = [
        'text/x-javascript',
        'text/javascript',
        'application/javascript'
    ]

    def _start_backend(self):
        
        self.backend.start(
            server.__file__,
            sys.executable,
            reuse=True,
            share_id='javascript'
        )

    @property
    def language(self):

        return 'javascript'

    def __repr__(self):

        return 'JavaScriptCodeEdit(path={})'.format(self.file.path)



# coding=utf-8
from libopensesame.py3compat import *
from libopensesame.exceptions import osexception
from libopensesame import python_workspace_api as py_api
import js2py.base


class Canvas(object):
    
    def __init__(self, kwargs={}):
        self._canvas = py_api.Canvas(**_cnv(kwargs))
    
    @property
    def color(self): return self._canvas.color
    @color.setter
    def color(self, val): self._canvas.color = val
    @property
    def background_color(self): return self._canvas.background_color
    @background_color.setter
    def background_color(self, val): self._canvas.background_color = val
    @property
    def fill(self): return self._canvas.fill
    @fill.setter
    def fill(self, val): self._canvas.fill = val
    @property
    def penwidth(self): return self._canvas.penwidth
    @penwidth.setter
    def penwidth(self, val): self._canvas.penwidth = val
    @property
    def html(self): return self._canvas.html
    @html.setter
    def html(self, val): self._canvas.html = val
    @property
    def font_family(self): return self._canvas.font_family
    @font_family.setter
    def font_family(self, val): self._canvas.font_family = val
    @property
    def font_size(self): return self._canvas.font_size
    @font_size.setter
    def font_size(self, val): self._canvas.font_size = val
    @property
    def font_italic(self): return self._canvas.font_italic
    @font_italic.setter
    def font_italic(self, val): self._canvas.font_italic = val
    @property
    def font_bold(self): return self._canvas.font_bold
    @font_bold.setter
    def font_bold(self, val): self._canvas.font_bold = val
    @property
    def font_underline(self): return self._canvas.font_underline
    @font_underline.setter
    def font_underline(self, val): self._canvas.font_underline = val
    
    def arrow(self, kwargs={}):
        self._canvas.arrow(**_args(kwargs, sx=0, sy=0, ex=50, ey=0))
        
    def clear(self, kwargs={}):
        self._canvas.clear(**_args(kwargs))
    
    def circle(self, kwargs={}):
        self._canvas.circle(**_args(kwargs, x=0, y=0, r=50))
        
    def ellipse(self, kwargs={}):
        self._canvas.ellipse(**_args(kwargs, x=-50, y=-25, w=100, h=50))
    
    def fixdot(self, kwargs={}):
        self._canvas.fixdot(**_args(kwargs))
        
    def gabor(self, kwargs={}):
        self._canvas.gabor(**_args(kwargs, x=0, y=0, orient=0, freq=.1))
        
    def image(self, kwargs={}):
        kwargs = _args(kwargs)
        if 'fname' not in kwargs:
            raise osexception(
                'fname is a required parameter for Canvas.image()')
        kwargs['fname'] = py_api.experiment.pool[kwargs['fname']]
        self._canvas.image(**kwargs)
        
    def line(self, kwargs={}):
        self._canvas.line(**_args(kwargs, sx=0, sy=0, ex=50, ey=0))
        
    def noise_patch(self, kwargs={}):
        self._canvas.noise_patch(**_args(kwargs, x=0, y=0))
    
    def polygon(self, kwargs={}):
        kwargs = _args(kwargs)
        if 'vertices' not in kwargs:
            raise osexception(
                'vertices is a required parameter for Canvas.polygon()')
        self._canvas.polygon(**kwargs)
        
    def rect(self, kwargs={}):
        self._canvas.rect(**_args(kwargs, x=-50, y=-25, w=100, h=50))
        
    def show(self):
        return self._canvas.show()
        
    def text(self, kwargs={}):
        kwargs = _args(kwargs)
        if 'text' not in kwargs:
            raise osexception(
                'text is a required parameter for Canvas.text()')
        self._canvas.text(**kwargs)
        
        
def xy_from_polar(rho, phi, pole=(0, 0)):
    return py_api.xy_from_polar(*_cnv_args(rho, phi, pole))
    
    
def xy_to_polar(x, y, pole=(0, 0)):
    return py_api.xy_to_polar(*_cnv_args(x, y, pole))
    
    
def xy_circle(n, rho, phi0=0, pole=(0, 0)):
    return py_api.xy_circle(*_cnv_args(n, rho, phi0, pole))
    
    
def xy_distance(x1, y1, x2, y2):
    return py_api.xy_distance(*_cnv_args(x1, y1, x2, y2))


def xy_grid(n, spacing, pole=(0, 0)):
    return py_api.xy_grid(*_cnv_args(n, spacing, pole))
    
    
def xy_random(n, width, height, min_dist=0, pole=(0, 0)):
    return py_api.xy_random(*_cnv_args(n, width, height, min_dist, pole))


def reset_feedback():
    py_api.reset_feedback()
    
    
def set_subject_nr(nr):
    py_api.set_subject_nr(_cnv(nr))
    
    
def sometimes(p=.5):
    py_api.sometimes(_cnv(p))
    
    
def _args(obj, **defaults):
    obj = _cnv(obj)
    for key, val in defaults.items():
        if key not in obj:
            obj[key] = val
    return obj


def _cnv_args(*args):
    
    return [
        _cnv(arg) for arg in args
        if not isinstance(arg, js2py.base.PyJsUndefined)
    ]


def _cnv(obj):
    if isinstance(obj, (js2py.base.PyJsNumber, js2py.base.PyJsString)):
        return obj.value
    if isinstance(obj, js2py.base.PyJsArray):
        return [_cnv(obj[val]) for val in obj]
    if isinstance(obj, js2py.base.JsObjectWrapper):
        try:
            return {key: _cnv(obj[key]) for key in obj}
        except TypeError:
            return [_cnv(val) for val in obj]
    return obj
